import math
import pandas as pd
from wallet.wallet import calculate_investment_value
from extra_print import print_green, print_red, \
    determine_print_color_from_prices, info_str


def simulator(time_range, traded_stocks, wallet, max_positions=5, take_profit=0, stop_loss=0, auto_traiding=False):
    """
    This function returns decorator for a stock market strategy.

    :param time_range: an array with traiding days only (no Saturdays, Sundays, holidays)
    :param traded_stocks: a dictionary with stocks.stock.Stock instances
    :param wallet: wallet.wallet.Wallet for traiding 
    :param max_positions: maximum number of different stocks in the wallet
    :param take_profit: if the price of a stock increases by this fraction, it will be sold;
        if equal 0 - take profit is deactivated (default)
    :param stop_loss: if the price of a stock decreases by this fraction (comparing to the 
        purchase price), it will be sold; if equal 0 - stop loss is deactivated (default)
    :param moving_stop_loss: if the price of a stock in a one day decreases by this fraction, 
        it will be sold; if equal 0 - stop loss is deactivated (default)
    :param auto_traiding: boolean, if True selling immediately when stop loss / take profit 
        is reached is simulated
    :return: decorator for a stock market strategy that returns change of wallet value 
        over given time_range
    """

    def strategy_wrapper(func):

        def simulation(*args, **kwargs):
            wallet_history = pd.DataFrame(columns=['Date', 'Wallet state'])
            stocks_to_buy = []
            stocks_to_sell = []

            for day in time_range:
                day_str = day.strftime('%Y-%m-%d')

                # Buy selected day before. Loop over list, order can be important here.
                # Strategy can sort relevant stocks - high priority first.
                for tck in stocks_to_buy:
                    if not wallet.get_volume_of_stocks(tck):
                        price = traded_stocks[tck].ohlc['Open'].get(day, None)
                        if price:
                            total = calculate_investment_value(wallet, max_positions)
                            total = total - wallet(total)  # needs some money to pay commission
                            volume = math.floor(total/price)
                            if volume > 0:
                                wallet.buy(tck, volume, price)
                                print(info_str(day_str, 'B', tck, volume, price))
                                # make sure to not sell it the same day
                                if tck in stocks_to_sell:
                                    stocks_to_sell.remove(tck)

                # Sell selected day before. List to set, we dont care here about the order.
                # Set will remove duplcates.
                for tck in set(stocks_to_sell):
                    if wallet.get_volume_of_stocks(tck):
                        price = traded_stocks[tck].ohlc['Open'].get(day, None)
                        if price: 
                            print_color = determine_print_color_from_prices(price, wallet.get_purchase_price_of_stocks(tck))
                            volume = wallet.sell_all(tck, price)
                            print_color(info_str(day_str, 'S', tck, volume, price))

                # call decorated function - strategy function
                stocks_to_buy, stocks_to_sell = func(day=day, traded_stocks=traded_stocks, *args, **kwargs)

                # if auto traiding is active, check if price of any stock in the wallet crossed
                # take proffit or stop loss price; if yes then sell it immediately 
                if auto_traiding:
                    for tck in wallet.list_stocks().copy():

                        price_max = traded_stocks[tck].ohlc['High'].get(day, None)
                        if price:
                            wallet.update_price(tck, price_max)
                        if wallet.change(tck) > take_profit:
                            price = wallet.get_purchase_price_of_stocks(tck) * (1+take_profit)  # selling it immediately
                            price = round(price, 2)
                            volume = wallet.sell_all(tck, price)
                            print_green(info_str(day_str, 'TP', tck, volume, price))
                            continue

                        price_min = traded_stocks[tck].ohlc['Low'].get(day, None)
                        if price:
                            wallet.update_price(tck, price_min)
                        if wallet.change(tck) < -stop_loss:
                            price = wallet.get_purchase_price_of_stocks(tck) * (1-stop_loss)  # selling it immediately
                            price = round(price, 2)
                            volume = wallet.sell_all(tck, price)
                            print_red(info_str(day_str, 'SL', tck, volume, price))
                            
                for tck in wallet.list_stocks():                        
                    # update the price to the closing price and calculate relative price change
                    ohlc = traded_stocks[tck].ohlc
                    price = ohlc['Close'].get(day, None)
                    if price:
                        wallet.update_price(tck, price)

                    # if auto traiding is not active, then take profit / stop loss the next day
                    if not auto_traiding:

                        # take profit the next day
                        if take_profit and wallet.change(tck) > take_profit:
                            stocks_to_sell.append(tck)

                        # stop loss the next day - price below purchase price
                        if stop_loss and wallet.change(tck) < -stop_loss:
                            stocks_to_sell.append(tck)

                # save history of the wallet
                wallet_history = wallet_history.append(
                    {'Date': day, 'Wallet state': wallet.total_value},
                    ignore_index=True
                    )
            
            return wallet_history

        return simulation

    return strategy_wrapper


if None:
    """
    This code should not be called - this is just a template function for tested strategy.
    """
    @simulator(TRAIDING_DAYS, stocks_data, MY_WALLET, TAKE_PROFIT, STOP_LOSS)
    def __strategy_template(arguments, *args, **kwargs):
        """
        :param arguments: any arguments needed for the strategy can be passed
        """
        day = kwargs['day']  # argument passed by decorator
        traded_stocks = kwargs['traded_stocks']  # argument passed by decorator
        stocks_to_buy = []
        stocks_to_sell = []

        """
        place for code that will fill in stocks_to_buy and stocks_to_sell sets with tickers
        """

        return stocks_to_buy, stocks_to_sell


if __name__ == '__main__':
    pass



import pandas as pd


def macd(prices: pd.DataFrame, mid_const: int=12, long_const: int=26, signal_const: int=9):
    """
    Returns Pandas Dataframe with MACD, Signal, and Histogram.

    :param prices:
    :param mid_const:
    :param long_const:
    :param signal_const:
    :returns:
    """
    price = prices['Close']
    output = pd.DataFrame(columns=['MACD', 'Signal', 'Histogram'])

    # calculate MACD line = price.emw(12) - price.emw(26)
    ewm_mid = price.ewm(span=mid_const).mean()
    ewm_long = price.ewm(span=long_const).mean()
    output['MACD'] = ewm_mid - ewm_long

    # calculate signal line = MACD.emw(9)
    output['Signal'] = output['MACD'].ewm(span=signal_const).mean()

    # calculate histogram = MACD - signal
    output['Histogram'] = output['MACD'] - output['Signal']
    
    return output


if __name__=='__main__':
    pass



import pandas as pd


def rsi(prices: pd.DataFrame, alpha: float=1/14):
    """
    Clculates Relative Strength Index (RSI).

    :prices: pandas.DataFrame with 'Close' column containing closing prices of stock
    :alpha: smoothing factor, 0 < alpha <= 1
    :return: pandas.Series with calculated RSI
    """
    price_now = prices['Close']
    price_prev = prices['Close'].shift(periods=1, fill_value=0)  # assign to each day price from the previous day

    price_changes = pd.DataFrame(columns=['Up', 'Down'])
    price_changes['Up'] = price_now - price_prev  # upward changes
    price_changes['Down'] = price_prev - price_now  # downward changes
    price_changes[price_changes < 0] = 0

    smma = price_changes.ewm(alpha = alpha, adjust=False).mean()  # smoothed moving averages
    rs = smma['Up'] / smma['Down']

    output_rsi = -100 / (rs+1)
    output_rsi = output_rsi + 100

    output_rsi = output_rsi.rename('RSI')
    output_rsi = output_rsi.to_frame()

    return output_rsi


def rsi_cross_signals(rsi_values: pd.DataFrame, cross_line: float, direction: str='onrise'):
    """
    Calculates buy/sell signals for given RSI signal line.

    :rsi_values: pandas.DataFrame with RSI column
    :cross_line: signal threshold line, when signal line crosses this line signal is set
    :direction: direction the signal line should cross threshold line 
        ('onrise' - signal increasing [default], 'onfall' - signal decreasing)
    :return: pandas.Series with with True values for days when signal appears
    """
    if not (0 < cross_line < 100):
        raise ValueError('cross_line takes values from 0 to 100')

    rsi_copy = pd.DataFrame()
    rsi_copy['RSI'] = rsi_values['RSI']
    rsi_copy['RSI day before'] = rsi_values['RSI'].shift(1)

    if 'onrise' == direction:
        # True signal if RSI is increasing and crossing the threshold line
        output = (rsi_copy['RSI'] >= cross_line ) & (rsi_copy['RSI day before'] < cross_line)
    elif 'onfall' == direction:
        # True signal if RSI is decreasing and crossing the threshold line
        output = (rsi_copy['RSI'] <= cross_line ) & (rsi_copy['RSI day before'] > cross_line)
    else:
        raise ValueError('wrong value for direction, must be "onrise" or "onfall"')

    return output



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="marketools",
    version="0.0.4",
    author="Albert Ratajczak",
    author_email="ratajczakalbert@gmail.com",
    description="Tools for stock market analysis.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AlbertRtk/marketools",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=[
        'pandas>=1.1.4',
        'requests>=2.25.0',
        'lxml>=4.6.2'
    ],
)



from .stqscraper import DATA_DIR
from .stqscraper.fundamentals import Fundamentals
from .stqscraper.stockquotes import StockQuotes
from .stock_index import wig20, mwig40
import pandas as pd
import os


class Stock:
    def __init__(self, ticker):
        self.ticker = ticker
        self._ohlc = StockQuotes(ticker)
        self._fundamentals = Fundamentals(ticker)

    # @property
    # def name(self):
    #     stocks = dict()
    #     return stocks.get(self.ticker, None)

    @property
    def ohlc(self):
        return self._ohlc.data

    @property
    def last_ohlc(self):
        return self.ohlc.iloc[-1]

    @property
    def date(self):
        """ Returns date of last price """
        output = self.last_ohlc.name
        output = output.strftime('%Y-%m-%d')
        return output

    @property
    def price(self):
        """ Returns last price """
        return self.last_ohlc['Close']

    @property
    def volume(self):
        """ Returns last session volume"""
        return self.last_ohlc['Volume']

    @property
    def fundamentals(self):
        if not bool(self._fundamentals):
            self._fundamentals.get_fundamentals()
        return self._fundamentals

    @property
    def eps(self):
        output = self.fundamentals.get('EPS')
        if output is not None:
            output = float(output)
        return output

    @property
    def pe(self):
        output = self.fundamentals.get('P/E')
        if output is not None:
            output = float(output)
        return output

    @property
    def pbv(self):
        output = self.fundamentals.get('P/BV')
        if output is not None:
            output = float(output)
        return output

    @property
    def dividend_yield(self):
        output = self.fundamentals.get('Dividend yield')
        if output is not None:
            output = output[0:-1]
            output = float(output)
        return output

    @property
    def stooq_plot_link(self):
        link = f'https://stooq.pl/q/a2/?s={self.ticker}'
        return link

    def mean_volume(self, over_last):
        volume = self.ohlc.tail(over_last)['Volume']
        output = volume.mean()
        return output


if __name__ == '__main__':
    pass



from . import DATA_DIR
from datetime import datetime
import pandas as pd
import requests
import os
import csv


class Fundamentals(dict):
    def __init__(self, ticker):
        super().__init__()
        self.ticker = ticker

    def get_fundamentals(self):
        update_required = True  # assuming that update will be required
        file_path = os.path.join(DATA_DIR, f'{self.ticker}_indicators.csv')

        if os.path.exists(file_path):
            timestamp_now = datetime.timestamp(datetime.now())
            timestamp_up = os.path.getatime(file_path)  # CSV file modification time

            if timestamp_now - timestamp_up < 24 * 3600:
                """ do not update more often than once in 24 hours """
                with open(file_path, 'r') as f:
                    reader = csv.DictReader(f)
                    try:
                        self.update(next(reader))
                    except StopIteration:
                        pass
                update_required = False

        if update_required:
            """ data older than 24 hours - update """
            self.update(self._download_fundamentals())
            with open(file_path, 'w') as f:
                writer = csv.DictWriter(f, self.keys())
                writer.writeheader()
                writer.writerow(self)

        if bool(self) is False:
            """ no fundamental data (None or empty dict) """
            self.update()

    def _download_fundamentals(self):
        """
        Gets available on stooq.com stock fundamentals for ticker.

        :return: 
        """
        """ URL to web page with fundamentals """
        url = f'https://stooq.pl/q/g/?s={self.ticker}'
        html = requests.get(url).text

        """ extracting table with fundamentals and setting value names as index """
        raw_table = pd.read_html(html)[0]
        idx = raw_table.iloc[:, 0]
        raw_table.set_index(idx, inplace=True)

        """ using extract_fundamentals_from_table to pars data """
        return self._extract_fundamental_from_df_to_dict(raw_table)

    @staticmethod
    def _extract_fundamental_from_df_to_dict(table):
        """
        Extracts stock fundamentals from table (Pandas DataFrame) and return them in dict.

        :param table: Pandas DataFrame with data from raw html table
        :return:
        """
        """ names of fundamental values (keys) and polish translations (values) - stooq is polish """
        fund_val = {'EPS': 'EPS (ttm)',
                    'P/E': 'C/Z (ttm)',
                    'P/BV': 'C/WK',
                    'Dividend yield': 'Stopa dywidendy'}

        """ creating and filling output dict """
        output = dict()
        for k in fund_val.keys():
            try:
                output[k] = table.loc[fund_val[k], 1]
            except KeyError:
                pass

        return output


if __name__ == '__main__':
    pass
